package main;

import java.io.FileNotFoundException;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import messif.algorithms.Algorithm;
import messif.algorithms.AlgorithmMethodException;
import messif.algorithms.impl.SequentialScan;
import messif.buckets.CapacityFullException;
import messif.buckets.impl.MemoryStorageBucket;
import messif.objects.LocalAbstractObject;
import messif.objects.impl.ObjectFloatVectorL2;
import messif.objects.util.AbstractObjectList;
import messif.operations.AnswerType;
import messif.operations.data.BulkInsertOperation;
import messif.operations.query.RangeQueryOperation;
import messif.statistics.OperationStatistics;

/**
 * Class to evaluate LAESA algorithm with the polynomial convex transform
 *
 * @author Vladimir Mic, xmic@fi.muni.cz
 */
public class Main {

    /**
     * Main class
     *
     * @param args paths to the dataset, pivots, queries, and the ground-truth.
     * All in the standard MESSIF format (Messif in the version 2)
     * @throws FileNotFoundException
     */
    public static void main(String[] args) throws FileNotFoundException {
        String datasetPath = args[0];
        String pivotFile = args[1];
        String queriesFile = args[2];
        // The ground truth is used to get the radii to the kth nearest neighbour.
        // It should contains answers of the kNN queries in a standard MESSIF format.
        //The distanes should be original with no onvex transform.
        String groundTruthPath = args[3];
        float exponent = 1.2f;

        PolynomialConvexTransform transform = new PolynomialConvexTransform(exponent, ObjectFloatVectorL2.class);
        Algorithm alg = buildAlgorithm(datasetPath, pivotFile, 256, transform.getClass());
        executeQueries(alg, queriesFile, groundTruthPath, transform.getClass(), 100, exponent);
    }

    /**
     *
     * @param datasetPath path to dataset (LocalAbstractObjects in the messif
     * txt format)
     * @param pivotsPath path to the pivot file (LocalAbstractObjects in the
     * messif txt format)
     * @param numberOfPivots number of pivots used for the LAESA algorithm
     * (first from the pivot file)
     * @param clazz class of the objects that extends the LocalAbstractObject
     * class
     * @return an instance of the LAESA algorithm with inserted all objects from
     * the dataset and precumputed distrances to the pivots.
     */
    private static Algorithm buildAlgorithm(String datasetPath, String pivotsPath, int numberOfPivots, Class clazz) {
        try {
            Algorithm ret = new SequentialScan(MemoryStorageBucket.class, IOTools.getIterator(clazz, pivotsPath), numberOfPivots, true);
            BulkInsertOperation insert = new BulkInsertOperation(IOTools.getObjects(clazz, datasetPath));
            ret.executeOperation(insert);
            return ret;
        } catch (AlgorithmMethodException | NoSuchMethodException | CapacityFullException | InstantiationException ex) {
            Logger.getLogger(Main.class.getName()).log(Level.SEVERE, null, ex);
        }
        return null;
    }

    /**
     *
     * @param alg an instance of the LAESA algorithm
     * @param queriesFile path to the query objects (LocalAbstractObjects in the
     * messif txt format)
     * @param groundTruthPath evaluated queries
     * @param clazz class of the objects that extends the LocalAbstractObject
     * class
     * @param k simulating the kNN queries as the range queries with the range
     * to the kth nearest neighbour found in the ground truth
     */
    private static void executeQueries(Algorithm alg, String queriesFile, String groundTruthPath, Class clazz, int k, float exponent) {
        Map<String, Float> radii = IOTools.parseRadiiForKNN(k, groundTruthPath);
        AbstractObjectList<LocalAbstractObject> queries = IOTools.getObjects(clazz, queriesFile);
        for (LocalAbstractObject queryObject : queries) {
            try {
                Float radius = radii.get(queryObject.getLocatorURI());
                radius = (float) Math.pow(radius, exponent);
                RangeQueryOperation query = new RangeQueryOperation(queryObject, radius, AnswerType.CLEARED_OBJECTS);
                alg.resetOperationStatistics();
                alg.statisticsBeforeOperation();
                query = alg.executeOperation(query);
                alg.statisticsAfterOperation(query);
                query.setParameter("DistanceComputations", OperationStatistics.getOpStatisticCounter("DistanceComputations").getValue());
                query.setParameter("DistanceComputations.Savings", OperationStatistics.getOpStatisticCounter("DistanceComputations.Savings").getValue());
                IOTools.printAnswer(query, System.out);
            } catch (NoSuchMethodException | AlgorithmMethodException ex) {
                Logger.getLogger(Main.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
    }
}
