package main;

import java.io.BufferedReader;
import java.io.EOFException;
import java.io.IOException;
import java.io.OutputStream;
import java.lang.reflect.InvocationTargetException;
import messif.objects.LocalAbstractObject;
import static messif.objects.LocalAbstractObject.MAX_DISTANCE;

/**
 * Wrapper for metric objects that perform polynomial convex transform of their
 * distances.
 *
 * @author Vladimir Mic, xmic@fi.muni.cz
 */
public class PolynomialConvexTransform extends LocalAbstractObject {

    private static float exp = 1f;
    private static float maxDistance = 1f;
    private LocalAbstractObject lao;
    private static Class clazz;
    public static int distsComputation = 0;

    /**
     *
     * @param exponent exponent of the polynomial convext transform
     * @param clazz class of the encapsulated metric object. For instance
     * ObjectFloatVectorL2 for Euclidean space with real-values vectors, or
     * MetaObjectSAPIRWeightedDist3 for MPEG7 descriptors.
     */
    public PolynomialConvexTransform(float exponent, Class clazz) {
        this(exponent, 1f, clazz);
    }

    /**
     *
     * @param exponent exponent of the polynomial convext transform
     * @param maxDistance maximum distance in the space. Use this constructor if
     * the polynomial convex transform leads to too big numbers (prevents
     * overflow of the distance)
     * @param clazz class of the encapsulated metric object. For instance
     * ObjectFloatVectorL2 for Euclidean space with real-values vectors, or
     * MetaObjectSAPIRWeightedDist3 for MPEG7 descriptors.
     */
    public PolynomialConvexTransform(float exponent, float maxDistance, Class clazz) {
        PolynomialConvexTransform.exp = exponent;
        PolynomialConvexTransform.maxDistance = maxDistance;
        PolynomialConvexTransform.clazz = clazz;
    }

    /**
     * Constructor used to deserialise LocalAbstractObjects from the text file
     * by the MESSIF library.
     *
     * @param br stream from the text file with metric objects (see class
     * LocalAbstractObject)
     * @throws Throwable throws expection for instance at the and of the file -
     * which is processed properly in the MESSIF library
     */
    public PolynomialConvexTransform(BufferedReader br) throws Throwable {
        try {
            lao = new TextStreamFactory(clazz).create(br);
        } catch (InvocationTargetException e) {
            if (e.getCause() instanceof EOFException) {
                throw e.getCause();
            }
        }
    }

    /**
     * For internal countingof the distance computations.
     */
    public static void resetDistComps() {
        distsComputation = 0;
    }

    @Override
    protected float getDistanceImpl(LocalAbstractObject obj, float distThreshold) {
        distsComputation++;
        PolynomialConvexTransform objCast = (PolynomialConvexTransform) obj;
        float origDist = lao.getDistance(objCast.lao) / maxDistance;
        if (origDist == MAX_DISTANCE) {
            return 1;
        }
        if (exp == 1f) {
            return origDist;
        }
        return (float) Math.pow(origDist, exp);
    }

    @Override
    public int getSize() {
        return lao.getSize();
    }

    @Override
    public boolean dataEquals(Object obj) {
        return lao.dataEquals(((PolynomialConvexTransform) obj).lao);
    }

    @Override
    public int dataHashCode() {
        return lao.dataHashCode();
    }

    @Override
    protected void writeData(OutputStream stream) throws IOException {
        lao.write(stream);
    }

    @Override
    public String getLocatorURI() {
        return lao.getLocatorURI();
    }
}
