/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package main.parser;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import main.IOTools;

/**
 * Parser of the results of the similarity search for the MESSIF library.
 *
 * @author Vladimir Mic, xmic@fi.muni.cz
 */
public class KNNResultFileParser {

    private final List<EvaluatedQuery> evaluatedQueries;

    /**
     * Constructor for parsing all the results from the text file.
     *
     * @param filePath path to the text file with the results
     */
    public KNNResultFileParser(String filePath) {
        this(filePath, -1);
    }

    /**
     *
     * @param filePath path to the text file with the results
     * @param k maximum number of the nearest neighbour taken for each query
     * objects
     */
    public KNNResultFileParser(String filePath, int k) {
        evaluatedQueries = new ArrayList<>();
        parse(filePath, k);
    }

    private void parse(String filePath, int k) {
        if (k < 0) {
            k = Integer.MAX_VALUE;
        }
        try {
            BufferedReader br = new BufferedReader(new FileReader(filePath));
            String locatorURI = null;
            List<String> nearestNeighbours = null;
            List<Float> distances = null;
            String firstLine = null;
            Integer time = null;
            Integer distanceComputations = null;
            try {
                while (true) {
                    String line = br.readLine();
                    if (line.isEmpty()) {
                        continue;
                    }
                    if (line.startsWith("OperationTime")) {
                        String timeS = line.substring(line.indexOf(" ") + 1);
                        time = Integer.parseInt(timeS);
                    }
                    if (line == null || line.contains("returned") || line.contains("query=")) {
                        if (locatorURI != null) {
                            EvaluatedQuery evaluatedQuery = new EvaluatedQuery(distanceComputations, firstLine, locatorURI, nearestNeighbours, distances, time);
                            if (!evaluatedQueries.contains(evaluatedQuery)) {
                                evaluatedQueries.add(evaluatedQuery);
                            } else {
                                Logger.getLogger(KNNResultFileParser.class.getName()).log(Level.WARNING, "Query evaluated again: {0}", new Object[]{locatorURI});
                            }
                        }
                        firstLine = line.substring(line.indexOf("was successful;") + "was successful;".length());
                        if (line.contains("returned")) {
                            locatorURI = line.substring(line.indexOf("(") + 1, line.indexOf(")"));
                        } else {
                            locatorURI = line.substring(line.indexOf("query=") + "query=".length());
                        }
                        nearestNeighbours = new ArrayList<>();
                        distances = new ArrayList<>();
                        time = null;
                    }
                    if (Character.isDigit(line.charAt(0))) {
                        String[] split = line.split(", ");
                        for (int i = 0; i < Math.min(k, split.length); i++) {
                            String[] pair = split[i].split(": ");
                            nearestNeighbours.add(pair[1]);
                            distances.add(Float.parseFloat(pair[0]));
                        }
                    }
                    if (line.contains("distance computations")) {
                        String distCompS = line.substring(line.indexOf("distance computations") + "distance computations".length() + 2);
                        distanceComputations = Integer.parseInt(distCompS);
                    }
                }
            } catch (NullPointerException ex) {
                evaluatedQueries.add(new EvaluatedQuery(distanceComputations, firstLine, locatorURI, nearestNeighbours, distances, time));
                if (!evaluatedQueries.isEmpty()) {
                    Logger.getLogger(KNNResultFileParser.class.getName()).log(Level.INFO, "Found {0} evaluated queries, first with {1} neighbours.", new Object[]{evaluatedQueries.size(), evaluatedQueries.get(0).nearestNeighbours.size()});
                }
            }
        } catch (IOException ex) {
            Logger.getLogger(KNNResultFileParser.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    public List<EvaluatedQuery> getEvaluatedQueries() {
        return evaluatedQueries;
    }

    public int getNumberOfEvaluatedQueries() {
        return evaluatedQueries.size();
    }

    /**
     * Structure for storing info about the evaluate similarity query
     */
    public static class EvaluatedQuery {

        private final String locatorURI;
        private final List<Integer> nearestNeighbours;
        private final List<Float> distances;
        private final Integer time;
        private final Integer distanceComputations;
        private final String firstLine;

        public EvaluatedQuery(Integer distanceComputations, String firstLine, String locatorURI, List<String> nearestNeighbours, List<Float> distances) {
            this(distanceComputations, firstLine, locatorURI, nearestNeighbours, distances, null);
        }

        public EvaluatedQuery(Integer distanceComputations, String firstLine, String locatorURI, List<String> nearestNeighbours, List<Float> distances, Integer time) {
            this.locatorURI = locatorURI;
            this.nearestNeighbours = IOTools.transformStringsToIntegers(nearestNeighbours);
            this.distances = distances;
            this.distanceComputations = distanceComputations;
            this.time = time;
            this.firstLine = firstLine;
        }

        public String getLocatorURI() {
            return locatorURI;
        }

        public List<Integer> getNearestNeighbours() {
            return nearestNeighbours;
        }

        public Set<Integer> getNearestNeighboursSet() {
            return new HashSet<>(nearestNeighbours);
        }

        public List<Float> getDistances() {
            return distances;
        }

        public String getFirstLine() {
            return firstLine;
        }

        public Integer getTime() {
            return time;
        }

        public Integer getDistanceComputations() {
            return distanceComputations;
        }

        @Override
        public int hashCode() {
            return locatorURI.hashCode();
        }

        @Override
        public boolean equals(Object obj
        ) {
            if (obj == null) {
                return false;
            }
            if (getClass() != obj.getClass()) {
                return false;
            }
            final EvaluatedQuery other = (EvaluatedQuery) obj;
            return locatorURI.equals(other.locatorURI);
        }

        public void print(String prefix, PrintStream out) {
            if (prefix == null) {
                out.println(firstLine);
            } else {
                out.println(prefix + locatorURI);
            }
            for (int i = 0; i < nearestNeighbours.size(); i++) {
                if (i > 0) {
                    out.print(", ");
                }
                Float dist = distances.get(i);
                Integer locator = nearestNeighbours.get(i);
                out.print(dist + ": " + locator);
            }
            out.println();
        }

    }

}
